import { createEffect, createRenderEffect } from "solid-js";
import { createRenderer } from "solid-js/universal";
import { createRoot, threeRenderer } from "./three/core";
import { Instance } from "./three/core/renderer";
import { log } from "./three/core/utils";

const PROPERTIES = new Set(["className", "class", "textContent"]);

const domRenderer = {
  createElement(el: string) {
    log("dom", "createElement", el);
    return document.createElement(el);
  },
  owns: function (node: any): node is HTMLElement {
    return node instanceof HTMLElement || node instanceof Text;
  },
  createTextNode(value) {
    log("dom", "createTextNode", value);
    return document.createTextNode(value);
  },
  replaceText(textNode: Text, value) {
    log("dom", "replaceText", textNode, value);
    textNode.data = value;
  },
  setProperty(node: HTMLElement, name: string, value: any) {
    log("dom", "setProperty", node, name, node[name], value);
    if (name === "style") Object.assign(node.style, value);
    if (name === "class") node.className = value;
    else if (name.startsWith("on")) node[name.toLowerCase()] = value;
    else if (PROPERTIES.has(name)) node[name] = value;
    else {
      node.setAttribute(name, value);
    }
  },
  insertNode(
    parent: HTMLElement,
    node: HTMLElement | Text,
    anchor?: HTMLElement | Text
  ) {
    log("dom", "insertNode", parent, node, anchor);
    if (
      parent instanceof HTMLElement &&
      (node instanceof HTMLElement || node instanceof Text)
    ) {
      parent.insertBefore(node, anchor);
    }
  },
  isTextNode(node) {
    log("dom", "isTextNode", node);
    return node.type === 3;
  },
  removeNode(parent, node) {
    log("dom", "removeNode", parent, node);
    if (parent instanceof HTMLElement) {
      parent.removeChild(node);
    } else {
      parent.remove(node);
    }
  },
  getParentNode(node) {
    return node.parentNode;
  },
  getFirstChild(node) {
    return node.firstChild;
  },
  getNextSibling(node) {
    return node.nextSibling;
  },
};

export const {
  render,
  effect,
  memo,
  createComponent,
  createElement,
  createTextNode,
  insertNode,
  insert,
  spread,
  setProp,
  mergeProps,
} = createRenderer<HTMLElement | Instance | Text>({
  createElement(string) {
    if (threeRenderer.knowsElement(string)) {
      return threeRenderer.createElement(string);
    } else {
      return domRenderer.createElement(string);
    }
  },
  createTextNode(value) {
    return domRenderer.createTextNode(value);
  },
  replaceText(textNode: Text, value: string) {
    return domRenderer.replaceText(textNode, value);
  },
  setProperty(node, name, value) {
    if (domRenderer.owns(node)) {
      return domRenderer.setProperty(node, name, value);
    }

    return threeRenderer.setProperty(node, name, value);
  },
  insertNode(parent, node, anchor) {
    if (domRenderer.owns(parent) && domRenderer.owns(node)) {
      return domRenderer.insertNode(parent, node, anchor as HTMLElement | Text);
    } else if (
      domRenderer.owns(parent) &&
      parent instanceof HTMLCanvasElement
    ) {
      let root = createRoot(parent, {});
      root.setState({ scene: node });
      // createRenderEffect(() => {
      // });
    }

    return threeRenderer.insertNode(parent, node, anchor);
  },
  isTextNode(node) {
    if (domRenderer.owns(node)) {
      return domRenderer.isTextNode(node);
    }
  },
  removeNode(parent, node) {
    if (domRenderer.owns(parent)) {
      return domRenderer.removeNode(parent, node);
    }
    return threeRenderer.removeNode(parent, node);
  },
  getParentNode(node) {
    if (domRenderer.owns(node)) {
      return domRenderer.getParentNode(node);
    }
    return threeRenderer.getParentNode(node);
  },
  getFirstChild(node) {
    if (domRenderer.owns(node)) {
      return domRenderer.getFirstChild(node);
    }
    return threeRenderer.getFirstChild(node);
  },
  getNextSibling(node) {
    return domRenderer.getNextSibling(node);
  },
});

export {
  For,
  Show,
  Suspense,
  SuspenseList,
  Switch,
  Match,
  Index,
  ErrorBoundary,
} from "solid-js";
