import * as React from "react";
import * as THREE from "three";
import { StoreApi as UseStore } from "zustand/vanilla";
import { pick, omit } from "../core/utils";
import { extend, createRoot, RenderProps } from "../core";
import { createPointerEvents } from "./events";
import { context, RootState } from "../core/store";
import { EventManager } from "../core/events";
import {
  createEffect,
  createSignal,
  ErrorBoundary,
  onMount,
  Show,
  Suspense,
} from "solid-js";
import createResizeObserver from "@solid-primitives/resize-observer";
import { Scene } from "three";

export interface Props
  extends Omit<RenderProps<HTMLCanvasElement>, "size" | "events">,
    React.HTMLAttributes<HTMLDivElement> {
  children: React.ReactNode;
  fallback?: React.ReactNode;
  // resize?: ResizeOptions;
  events?: (store: UseStore<RootState>) => EventManager<any>;
}

type SetBlock = false | Promise<null> | null;
type UnblockProps = {
  set: React.Dispatch<React.SetStateAction<SetBlock>>;
  children: React.ReactNode;
};

extend(THREE);

const CANVAS_PROPS: Array<keyof Props> = [
  "gl",
  "events",
  "shadows",
  "linear",
  "flat",
  "orthographic",
  "frameloop",
  "dpr",
  "performance",
  "clock",
  "raycaster",
  "camera",
  "onPointerMissed",
  "onCreated",
];

function Block({ set }: Omit<UnblockProps, "children">) {
  React.useLayoutEffect(() => {
    set(new Promise(() => null));
    return () => set(false);
  }, [set]);
  return null;
}

// export const Canvas = function Canvas(props) {
//   // Create a known catalogue of Threejs-native elements
//   // This will include the entire THREE namespace by default, users can extend
//   // their own elements by using the createRoot API instead
//   React.useMemo(() => extend(THREE), []);

//   const [containerRef, { width, height }] = useMeasure({
//     scroll: true,
//     debounce: { scroll: 50, resize: 0 },
//     ...resize,
//   });
//   let canvasRef: HTMLCanvasElement;

//   const canvasProps = pick<Props>(props, CANVAS_PROPS);
//   const divProps = omit<Props>(props, CANVAS_PROPS);
// const [block, setBlock] = React.useState<SetBlock>(false);
// const [error, setError] = React.useState<any>(false);

// Suspend this component if block is a promise (2nd run)
// if (block) throw block;
// Throw exception outwards if anything within canvas throws
// if (error) throw error;

// if (width > 0 && height > 0 && canvas) {
//   createRoot(canvas, {
//     ...canvasProps,
//     size: { width, height },
//     events: events || createPointerEvents,
//   }).render();
// }

// React.useLayoutEffect(() => {
//   setCanvas(canvasRef.current);
// }, []);

// React.useEffect(() => {
//   return () => unmountComponentAtNode(canvas!);
// }, [canvas]);

export function Canvas(props) {
  let canvas: HTMLCanvasElement;
  let containerRef: HTMLDivElement;

  createEffect(() => {
    const root = createRoot(canvas, {
      events: createPointerEvents,
      size: containerRef.getBoundingClientRect(),
      camera: props.camera,
    });

    new ResizeObserver((entries) => {
      if (entries[0]?.target !== containerRef) return;
      root
        .getState()
        .setSize(entries[0].contentRect.width, entries[0].contentRect.height);
    }).observe(containerRef);

    root.setState({
      scene: (
        (
          <context.Provider value={root}>
            <scene>{props.children}</scene>
          </context.Provider>
        ) as unknown as Accessor<Scene>
      )(),
    });
  });

  return (
    <div
      id={props.id}
      className={`h-full w-full relative overflow-hidden ${
        props.className ?? ""
      }`}
      tabIndex={props.tabIndex}
      style={props.style}
      ref={containerRef}
    >
      <canvas class="h-full w-full" ref={canvas} />
    </div>
  );
}
